program ycp;
  uses UniLEX (*){*), UniPascal, BIOS_CRT {*};

  {$I GBL_Y_FL.pas }                                {   Y code file }

  var Nmodules,                                       {     }
      Nparas:   natural;                        {   () }
      InfoSize: longint;                      {     interface }
      CodeSize: longint;                 {       }
      BlockSize: natural;

      temp_name,                                         { temporary filename }
      in_name,                                               { input filename }
      out_name: string;                                     { output filename }
      in_file: file;                                             { input file }
      out_file: file;                                           { output file }
      plural: array [boolean] of string[1];        { constant array = '', 's' }
      errorsize,
      inited,                                {     }
      interrupted: boolean;                                 {    }
      CFH: CFHeader;
      backstr: string[31];

                                                               {$ifdef UniPas }
  procedure rename(const old, new: string); code $5e, $3b;

  procedure remove(s: string);              code $5e, $37;             {$endif}

  procedure TestRead;
  begin
    if IOresult <> 0 then begin
      writeln(message);
      writeln(message, 'Error reading file ', in_name);
      halt(7);
    end { if };
  end { TestRead };

  procedure TestWrite;
  begin
    if IOresult <> 0 then begin
      writeln(message);
      writeln(message, 'Error writing resulting file.');
      halt(7);
    end { if };
  end { TestRead };

  procedure ReadNbytes(pos: LongInt; var Buff; N: natural);
  begin
    if 0 <= pos then begin Seek(in_file, pos); TestRead;end;
    BlockRead(in_file, Buff, N); TestRead;
  end { ReadNbytes };

  procedure WriteNbytes(pos: LongInt; var Buff; N: natural);
  begin
    if 0 <= pos then begin Seek(out_file, pos); TestWrite; end;
    BlockWrite(out_file, Buff, N); TestWrite;
  end { WriteNbytes };

  procedure WriteZeros(pos: LongInt; N: natural);
    var i: natural;
        buff: packed array [0..1023] of byte;                  { 1   }
  begin
    fillchar(buff, sizeof(buff), #0);
    if 0 <= pos then Seek(out_file, pos); TestWrite;
    i:= sizeof(buff);
    repeat if i > N then i:= N;
      BlockWrite(out_file, Buff, i); TestWrite;
      dec(N, i);
    until N = 0;
  end { WriteZeros };

  procedure initialize;

    procedure helpuser;
    begin { helpuser }
      writeln;
      writeln('usage: YCP filename[.pgm] [new_name] [/blocksize]');
      halt(1);
    end { helpuser };

    procedure CheckCodeFile;
    begin
      ReadNbytes(0, CFH, SizeOf(CFH));
      if swap(CFH.magic) <> $c0de then begin
        writeln(message);
        writeln(message, in_name, ' is not a code file!'); halt(7);
      end { if };
      if (CFH.Wsz <> 16) or (CFH.Bsx <> 0) then begin
        writeln(message);
        writeln(message, in_name, 'has invalid word size(', CFH.Wsz,
                               ') or byte sex(', CFH.Bsx, ')');
        halt(7);
      end { if };
      if 1 < CFH.VerMaj then begin
        writeln(message);
        writeln(message, in_name, ' can''t handle');
        writeln(message, 'version ', CFH.VerMaj, '.', CFH.VerMin);
        halt(7);
      end { if };

      if (CFH.ParaSz <> 32) (* or (CFH.BlckSz <> 512) *) then begin
        writeln(message); writeln(message, 'Can''t handle ', in_name, '!');
        writeln(message, 'Paragraph size is ', CFH.ParaSz,
                         ' block size ', CFH.BlckSz);
        halt(7);
      end { if };
      writeln('Processing file ', in_name, ', current block size is ',
              CFH.BlckSz, ' byte', plural[CFH.BlckSz <> 1], '.');
    end { CheckCodeFile };

    procedure Extension(var name: string; var ext: string(*)[4](*));
      var i, j, k: integer;
    begin  {   ( )   ( ) .ext   }
      i:= length(name);     {        .  \ }
      j:= i +ScanFor('\', name, i, -i);        {       }
      k:= j +1 +ScanFor('.', name, j+1, i - j); {   () }

      if i < k then                               {      }
        insert(ext, name, i+1)                      {    ext }
      else begin
        j:= i -k +1;                    {      ext }
        if j > 4 then j:= 4;
        move(name[k], ext[1], j); ext[0]:= chr(j);    {   ext }
      end { if };
    end { Extension };

    procedure get_path((*)const(*) name: string; var path: string(*)[80](*));
      var i, j, k: integer;
    begin                                 {      }
      i:= length(name);     {        \  : }
      j:= i +ScanFor('\', name, i, -i);        {       }
      if j = 0 then j:= pos(':', name);      {    \   : }
      move(name, path, j+1);
      path[0]:= chr(j);                     { j      :  \ }
    end { get_path };

    var endpos, wrk, para_no: cardinal;
        s: string;

  begin { initialize }
    writeln(output, 'Y Code Pack. version 1.00.' (*): -40 (*),
                    ' (c) 1992 Software R&D Lab., Sofia.': 39);

    Plural[false]:= ''; Plural[true]:= 's';
    fillchar(backstr, sizeof(backstr), #8);
    in_name:= ''; out_name:= ''; BlockSize:= 0;
    if paramcount = 0 then begin
      helpuser;
    end { if };
    for para_no:= 1 to ParamCount do begin
      ParamStr(para_no, s);
      if s[1] = '/' then begin
                      {       Block Size }
        delete(s, 1, 1); BlockSize:= GetInteger(s, 1, length(s), endpos);
        if (EndPos <> length(s)) or (IOresult <> 0) then begin
          writeln('*** Invalid parameter: /', s); helpuser;
        end { if };
                                 {       < 16 }
        endpos:= BlockSize; wrk:= 0;
        while not odd(endpos) and (endpos <> 0) and (wrk <= 16) do begin
          endpos:= endpos div 2; inc(wrk);
        end { while };
                             {           }
        if (endpos <> 1) or (wrk >= 16) then begin
          writeln('*** Invalid block size: /', s); helpuser;
        end { if }; end
      else if in_name = '' then
        in_name:= s
      else if out_name = '' then
        out_name:= s
      else begin
        writeln('*** Error! Too many parameters!'); HelpUser;
      end { if };
    end { for };

    CodeSize:= 0; InfoSize:= 0; Nparas:= 0; Nmodules:= 0;

    if in_name = '' then begin        {       }
      writeln('No file specified!'); helpuser;
    end { if };

    s:= '.pgm';
    Extension(in_name, s);            {  ( )  ext }

    if out_name = '' then begin {        }
      out_name:= in_name;
    end { if };

    Extension(out_name, s);       {   ext    in_name }

         {    (  )      }
    (*)
        reset(in_file, in_name);
    {*)
        FileMode:= 0;
        assign(in_file, in_name); reset(in_file,1);
        FileMode:= 2;
    {*}

    if IOresult <> 0 then begin
      writeln('Can''t access file: ', in_name); halt(1);
    end { if };

    CheckCodeFile;                           {    Y code  }

      {   (     )     }
             {          ..    }
    get_path(out_name, temp_name);

    (*)
        rewrite(out_file, temp_name, temporary);
    {*)
        insert('tempfile.ycp', temp_name, length(temp_name)+1);
        assign(out_file, temp_name); rewrite(out_file, 1);
    {*}

    if IOresult <> 0 then begin
      writeln('Can''t create new file!'); halt(7);
    end { if };
  end { initialize };

  procedure FinishUp;

    procedure err_create;
    begin { err_create }
      writeln;
      writeln('Can''t create new file: ', out_name); halt(7);
      if temp_name <> '' then begin
        remove(temp_name);
      end { if };
    end { err_create };

  begin { FinishUp }
    if inited and not errorsize and not interrupted then begin
      close(out_file);             if IOresult <> 0 then err_create;
      remove(out_name);    (*) {*) if IOresult <> 0 then {OK no old file{*};
      rename(temp_name, out_name); if IOresult <> 0 then err_create;
      writeln('File processed. New block size is ', BlockSize,
              ' byte', plural[BlockSize <> 1], '.'); end
    else if inited and (temp_name <> '') then begin
      close(out_file); remove(temp_name);
    end { if };
  end { FinishUp };

  procedure calc_size;
    var module: cardinal;
        wrk: cardinal;
        MD: ModDescriptor;
        UD: UsageDescriptor;
        SD: SegDescriptor;
  begin { calc_size }
    {     CFH     .    }
    {            .          }
    inc(Nparas);                  {   ( 0)    }
    module:= CFH.Fmodule;                            {    }
    while module <> 0 do begin              {      }
      ReadNbytes(longint(module) * para_size, MD, sizeof(MD));
      write(message, #13'Module: ', MD.name (*), #24); {*)); clreol; {*}
      inc(Nparas);        {     (   ) }
      inc(Nmodules);                                        {   }
      module:= MD.next;        {        }

      inc(InfoSize, MD.CinfoSz); {   interface  ( ) }

      wrk:= MD.UsageP;                   {    Usage  }
      backstr[0]:= chr(6 + 8);
      while wrk <> 0 do begin
        ReadNbytes(longint(wrk) * para_size, UD, sizeof(UD));
        (*){*)clreol;{*} write(message, (*)#24(*)' uses ', UD.name, backstr);
        inc(Nparas); wrk:= UD.next;
      end { while };
      (*) write(message, #24); {*) clreol {*};

      wrk:= MD.Mseg;                            {     }
      backstr[0]:= chr(9 + 8);
      while wrk <> 0 do begin
        ReadNbytes(longint(wrk) * para_size, SD, sizeof(SD));
        (*){*)clreol;{*} write(message, (*)#24(*)' segment ', SD.name, backstr);
        inc(Nparas);
        inc(CodeSize, SD.size);                      {    }
        wrk:= SD.nextseg;
      end { while };
      (*) write(message, #24); {*) clreol {*};
                {             }
    end { while };
    write(message, #13(*)#24);{*)); clreol; {*}
    writeln('File contains:', Nmodules:4, ' module', plural[Nmodules<>1], ',');
    writeln(                Nparas:   18, ' paragraph descriptors,');
    writeln(                InfoSize: 18, ' bytes interfaced information,');
    writeln(                CodeSize: 18, ' bytes Y code.');
  end { calc_size };

  procedure calc_new_size;
    var wrk, calculated: integer;
        total:      longint;
  begin
    total:= CodeSize + InfoSize + longint(Nparas) * para_size;
    wrk:= total div 65536 + ord(total mod 65536 <> 0);   {   , }
    {         2,   .  }

    calculated:= 1;                     {     2 }
    while calculated < wrk do begin          {  -    }
      calculated:= calculated *2;
    end { while };

    if BlockSize = 0 then           {        }
      BlockSize:= calculated
    else begin
      writeln('Minimal block size is: ', calculated, '.');
      if BlockSize < calculated then begin
        writeln('Error! Requested block size is: ', BlockSize, '.');
        errorsize:= true;
      end { if };
    end { if };
  end { calc_new_size };

  procedure copy_file;
    var NextModPara,          {       }
        NextFreePara: natural;                  {    }
        out_MOD,                   {       }
        in_MOD: natural;                            {     }
        data_offset: longint;
        newCFH: CFHeader;
        cur_MD, prev_MD: ModDescriptor;

    procedure AllignOnBlockBoundary;
      var wrk: natural;
    begin { AllignOnBlockBoundary }
      wrk:= data_offset mod BlockSize;
      if wrk <> 0 then begin    {      ,   }
        wrk:= BlockSize - wrk;   {          }
        WriteZeros(data_offset, wrk);{      0 }
        inc(data_offset, wrk);
      end { if };
    end { AllignOnBlockBoundary };

    function copy_data(in_Block, size: cardinal): cardinal;
      type ba = packed array [0..1023] of byte;
      var in_pos, out_pos: longint;
          n, buffsz: cardinal;
          buffer: ^ba;
    begin { copy_data }
      {    in_file  out_file,  in_block  offset  , }
      {      in_file.     data_offset }
      {       Block boundary,    . }
      AllignOnBlockBoundary; copy_data:= data_offset div BlockSize;
      in_pos:= CFH.BlckSz * longint(in_Block); out_pos:= data_offset;
      if size < MaxAvail (*)*2(*) - longint(1024) then      {   }
        buffsz:= size                                       {    }
      else begin                                                    {   }
        buffsz:= MaxAvail (*)*2(*) - longint(1024); {     }
      end { if };
      if buffsz > 32767 then buffsz:= 32000;
      getmemwords(pointer(buffer), (buffsz+1) div 2);      {    }
      repeat n:= buffsz;
        if n > size then n:= size;
        ReadNbytes(in_pos, buffer^, n);   in_pos:= -1;  { don't seek any more }
        WriteNbytes(out_pos, buffer^, n); out_pos:= -1; { don't seek any more }
        dec(size, n); inc(data_offset, n);
      until size = 0;
      freememwords(buffer, (buffsz+1) div 2);{     }
    end { copy_data };

    procedure copy_module(var in_MOD: natural; var MD: ModDescriptor);

      procedure copy_usagelst(var para: cardinal);
        var wrk: cardinal; UD: UsageDescriptor;
      begin { copy_usagelst }
        wrk:= para; para:= 0;               {    Usage  }
        backstr[0]:= chr(6 + 8);     {   ,    }
        while wrk <> 0 do begin
          ReadNbytes(longint(wrk) * para_size, UD, sizeof(UD));
          (*){*)clreol;{*} write(message, (*)#24(*)' uses ', UD.name, backstr);
          if para = 0 then para:= NextFreePara;
          wrk:= UD.next; inc(NextFreePara);
          if wrk <> 0 then begin
            UD.next:= NextFreePara;
          end { if };
          WriteNbytes(longint(NextFreePara-1) * para_size, UD, sizeof(UD));
        end { while };
        (*) write(message, #24); {*) clreol {*};
      end { copy_usagelst };

      procedure copy_segments(var para: cardinal);
        var wrk, new_blck: cardinal;
            SD: SegDescriptor;
      begin { copy_segments }
        wrk:= para; para:= 0;                   {     }
        backstr[0]:= chr(9 + 8);     {   ,    }
        while wrk <> 0 do begin
          ReadNbytes(longint(wrk) * para_size, SD, sizeof(SD));
          (*){*)clreol;{*} write(message,(*)#24(*)' segment ',SD.name,backstr);
          if para = 0 then para:= NextFreePara;
          wrk:= SD.nextseg; inc(NextFreePara);
          if wrk <> 0 then begin
            SD.nextseg:= NextFreePara;
          end { if };

          {    Usage       Y code. }
          SD.Start:= copy_data(SD.Start, SD.Size);            { new block pos }
          WriteNbytes(longint(NextFreePara-1) * para_size, SD, sizeof(SD));
        end { while };
        (*) write(message, #24); {*) clreol {*};
      end { copy_segments };

    begin { copy_module }
      ReadNbytes(longint(in_MOD) * para_size, MD, sizeof(MD));
      write(message, #13'Copying module: ', MD.name (*), #24);{*)); clreol; {*}

      copy_usagelst(MD.UsageP);     {      USES }

      copy_segments(MD.Mseg);                      {    }

      {    }
      if MD.CinfoSt <> 0 then begin
        backstr[0]:= #10;
        (*){*)clreol;{*} write(message, (*)#24(*)' interface', backstr);
        MD.CinfoSt:= copy_data(MD.CinfoSt, MD.CinfoSz);       { new block pos }
      end { if };
      write(message, #13(*)#24);{*)); clreol; {*}
    end { copy_module };

  begin { copy_file }
    write(message, 'Creating new file with block size ', BlockSize, '.');
    {            . }
    {       . }

    newCFH:= CFH;                                  {   file header- }
    with newCFH do begin
      vermin:= vermin and byte($7f);     {    ( CheckSum)}
      BlckSz:= BlockSize;                         {    BlockSize }
      Fmodule:= 0;                         {  ,      }
    end { with };
    NextModPara:= 1;         {         }
    NextFreePara:= Nmodules +1;       {       }
    data_offset:= Nparas * para_size; {        }

    out_MOD:= 0;
    in_MOD:= CFH.Fmodule;         {       }
    while in_MOD <> 0 do begin
      copy_module(in_MOD, cur_MD); in_MOD:= cur_MD.next;  {   }
      if out_MOD = 0 then
        newCFH.Fmodule:= NextModPara
      else begin
        prev_MD.next:= NextModPara;   {      }
        WriteNbytes(longint(out_mod) * para_size, prev_MD, SizeOf(prev_MD));
      end { if };
      out_mod:= NextModPara; inc(NextModPara); prev_MD:= cur_MD;
    end { while };
    if out_mod = 0 then begin
      writeln('*** Empty file. No modules in file!'); halt(7); end
    else begin
      prev_MD.next:= 0;{       (     ) }
      WriteNbytes(longint(out_mod) * para_size, prev_MD, SizeOf(prev_MD));
    end { if };
    WriteNbytes(0, newCFH, SizeOf(newCFH));         { finaly write the header }
  end { copy_file };

begin { ycp }
  interrupted:= true; inited:= false; temp_name:= ''; errorsize:= false;
  initialize;
  calc_size;
  calc_new_size;
  copy_file;
  interrupted:= false; inited:= true;                                       (*)
exit:                                                                       (*)
  FinishUp;
end { ycp }.